@@ -51,15 +51,16 @@ void Unpack::DoUnpack(int Method,bool Solid)
   switch(Method)
   {
 #ifndef SFX_MODULE
-    case 15:
+    case 15: // rar 1.5 compression
       Unpack15(Solid);
       break;
-    case 20:
-    case 26:
+    case 20: // rar 2.x compression
+    case 26: // files larger than 2GB
       Unpack20(Solid);
       break;
 #endif
-    case 29:
+    case 29: // rar 3.x compression
+    case 36: // alternative hash
       Unpack29(Solid);
       break;
   }
@@ -491,28 +492,35 @@ bool Unpack::AddVMCode(unsigned int FirstByte,byte *Code,int CodeSize)
       FiltPos--;
   }
   else
-    FiltPos=LastFilter;
+    FiltPos=LastFilter; // use the same filter as last time
+
   if (FiltPos>Filters.Size() || FiltPos>OldFilterLengths.Size())
     return(false);
   LastFilter=FiltPos;
   bool NewFilter=(FiltPos==Filters.Size());
 
+  UnpackFilter *StackFilter=new UnpackFilter; // new filter for PrgStack
+
   UnpackFilter *Filter;
-  if (NewFilter)
+  if (NewFilter) // new filter code, never used before since VM reset
   {
+    // too many different filters, corrupt archive
+    if (FiltPos>1024)
+      return(false);
+
     Filters.Add(1);
     Filters[Filters.Size()-1]=Filter=new UnpackFilter;
+    StackFilter->ParentFilter=Filters.Size()-1;
     OldFilterLengths.Add(1);
     Filter->ExecCount=0;
   }
-  else
+  else  // filter was used in the past
   {
     Filter=Filters[FiltPos];
+    StackFilter->ParentFilter=FiltPos;
     Filter->ExecCount++;
   }
 
-  UnpackFilter *StackFilter=new UnpackFilter;
-
   int EmptyCount=0;
   for (int I=0;I<PrgStack.Size();I++)
   {
@@ -549,7 +557,8 @@ bool Unpack::AddVMCode(unsigned int FirstByte,byte *Code,int CodeSize)
   StackFilter->Prg.InitR[3]=VM_GLOBALMEMADDR;
   StackFilter->Prg.InitR[4]=StackFilter->BlockLength;
   StackFilter->Prg.InitR[5]=StackFilter->ExecCount;
-  if (FirstByte & 0x10)
+
+  if (FirstByte & 0x10)   // set registers to optional parameters if any
   {
     unsigned int InitMask=Inp.fgetbits()>>9;
     Inp.faddbits(7);
@@ -557,6 +566,7 @@ bool Unpack::AddVMCode(unsigned int FirstByte,byte *Code,int CodeSize)
       if (InitMask & (1<<I))
         StackFilter->Prg.InitR[I]=RarVM::ReadData(Inp);
   }
+
   if (NewFilter)
   {
     uint VMCodeSize=RarVM::ReadData(Inp);
@@ -565,6 +575,8 @@ bool Unpack::AddVMCode(unsigned int FirstByte,byte *Code,int CodeSize)
     Array<byte> VMCode(VMCodeSize);
     for (int I=0;I<VMCodeSize;I++)
     {
+      if (Inp.Overflow(3))
+        return(false);
       VMCode[I]=Inp.fgetbits()>>8;
       Inp.faddbits(8);
     }
@@ -576,6 +588,7 @@ bool Unpack::AddVMCode(unsigned int FirstByte,byte *Code,int CodeSize)
   int StaticDataSize=Filter->Prg.StaticData.Size();
   if (StaticDataSize>0 && StaticDataSize<VM_GLOBALMEMSIZE)
   {
+    // read statically defined data contained in DB commands
     StackFilter->Prg.StaticData.Add(StaticDataSize);
     memcpy(&StackFilter->Prg.StaticData[0],&Filter->Prg.StaticData[0],StaticDataSize);
   }
@@ -587,16 +600,18 @@ bool Unpack::AddVMCode(unsigned int FirstByte,byte *Code,int CodeSize)
   }
   byte *GlobalData=&StackFilter->Prg.GlobalData[0];
   for (int I=0;I<7;I++)
-    VM.SetValue((uint *)&GlobalData[I*4],StackFilter->Prg.InitR[I]);
-  VM.SetValue((uint *)&GlobalData[0x1c],StackFilter->BlockLength);
-  VM.SetValue((uint *)&GlobalData[0x20],0);
-  VM.SetValue((uint *)&GlobalData[0x2c],StackFilter->ExecCount);
+    VM.SetLowEndianValue((uint *)&GlobalData[I*4],StackFilter->Prg.InitR[I]);
+  VM.SetLowEndianValue((uint *)&GlobalData[0x1c],StackFilter->BlockLength);
+  VM.SetLowEndianValue((uint *)&GlobalData[0x20],0);
+  VM.SetLowEndianValue((uint *)&GlobalData[0x2c],StackFilter->ExecCount);
   memset(&GlobalData[0x30],0,16);
 
-  if (FirstByte & 8)
+  if (FirstByte & 8) // put data block passed as parameter if any
   {
+    if (Inp.Overflow(3))
+      return(false);
     uint DataSize=RarVM::ReadData(Inp);
-    if (DataSize>=0x10000)
+    if (DataSize>VM_GLOBALMEMSIZE-VM_FIXEDGLOBALSIZE)
       return(false);
     unsigned int CurSize=StackFilter->Prg.GlobalData.Size();
     if (CurSize<DataSize+VM_FIXEDGLOBALSIZE)
@@ -604,6 +619,8 @@ bool Unpack::AddVMCode(unsigned int FirstByte,byte *Code,int CodeSize)
     byte *GlobalData=&StackFilter->Prg.GlobalData[VM_FIXEDGLOBALSIZE];
     for (int I=0;I<DataSize;I++)
     {
+      if (Inp.Overflow(3))
+        return(false);
       GlobalData[I]=Inp.fgetbits()>>8;
       Inp.faddbits(8);
     }
@@ -669,9 +686,29 @@ void Unpack::UnpWriteBuf()
           VM.SetMemory(0,Window+BlockStart,FirstPartLength);
           VM.SetMemory(FirstPartLength,Window,BlockEnd);
         }
+
+        VM_PreparedProgram *ParentPrg=&Filters[flt->ParentFilter]->Prg;
         VM_PreparedProgram *Prg=&flt->Prg;
+
+        if (ParentPrg->GlobalData.Size()>VM_FIXEDGLOBALSIZE)
+        {
+          // copy global data from previous script execution if any
+          Prg->GlobalData.Alloc(ParentPrg->GlobalData.Size());
+          memcpy(&Prg->GlobalData[VM_FIXEDGLOBALSIZE],&ParentPrg->GlobalData[VM_FIXEDGLOBALSIZE],ParentPrg->GlobalData.Size()-VM_FIXEDGLOBALSIZE);
+        }
+
         ExecuteCode(Prg);
 
+        if (Prg->GlobalData.Size()>VM_FIXEDGLOBALSIZE)
+        {
+          // save global data for next script execution
+          if (ParentPrg->GlobalData.Size()<Prg->GlobalData.Size())
+            ParentPrg->GlobalData.Alloc(Prg->GlobalData.Size());
+          memcpy(&ParentPrg->GlobalData[VM_FIXEDGLOBALSIZE],&Prg->GlobalData[VM_FIXEDGLOBALSIZE],Prg->GlobalData.Size()-VM_FIXEDGLOBALSIZE);
+        }
+        else
+          ParentPrg->GlobalData.Reset();
+
         byte *FilteredData=Prg->FilteredData;
         unsigned int FilteredDataSize=Prg->FilteredDataSize;
 
@@ -683,9 +720,33 @@ void Unpack::UnpWriteBuf()
           if (NextFilter==NULL || NextFilter->BlockStart!=BlockStart ||
               NextFilter->BlockLength!=FilteredDataSize || NextFilter->NextWindow)
             break;
+
+          // apply several filters to same data block
+
           VM.SetMemory(0,FilteredData,FilteredDataSize);
-          VM_PreparedProgram *NextPrg=&PrgStack[I+1]->Prg;
+
+          VM_PreparedProgram *ParentPrg=&Filters[NextFilter->ParentFilter]->Prg;
+          VM_PreparedProgram *NextPrg=&NextFilter->Prg;
+
+          if (ParentPrg->GlobalData.Size()>VM_FIXEDGLOBALSIZE)
+          {
+            // copy global data from previous script execution if any
+            NextPrg->GlobalData.Alloc(ParentPrg->GlobalData.Size());
+            memcpy(&NextPrg->GlobalData[VM_FIXEDGLOBALSIZE],&ParentPrg->GlobalData[VM_FIXEDGLOBALSIZE],ParentPrg->GlobalData.Size()-VM_FIXEDGLOBALSIZE);
+          }
+
           ExecuteCode(NextPrg);
+
+          if (NextPrg->GlobalData.Size()>VM_FIXEDGLOBALSIZE)
+          {
+            // save global data for next script execution
+            if (ParentPrg->GlobalData.Size()<NextPrg->GlobalData.Size())
+              ParentPrg->GlobalData.Alloc(NextPrg->GlobalData.Size());
+            memcpy(&ParentPrg->GlobalData[VM_FIXEDGLOBALSIZE],&NextPrg->GlobalData[VM_FIXEDGLOBALSIZE],NextPrg->GlobalData.Size()-VM_FIXEDGLOBALSIZE);
+          }
+          else
+            ParentPrg->GlobalData.Reset();
+
           FilteredData=NextPrg->FilteredData;
           FilteredDataSize=NextPrg->FilteredDataSize;
           I++;
@@ -722,8 +783,8 @@ void Unpack::ExecuteCode(VM_PreparedProgram *Prg)
   if (Prg->GlobalData.Size()>0)
   {
     Prg->InitR[6]=int64to32(WrittenFileSize);
-    VM.SetValue((uint *)&Prg->GlobalData[0x24],int64to32(WrittenFileSize));
-    VM.SetValue((uint *)&Prg->GlobalData[0x28],int64to32(WrittenFileSize>>32));
+    VM.SetLowEndianValue((uint *)&Prg->GlobalData[0x24],int64to32(WrittenFileSize));
+    VM.SetLowEndianValue((uint *)&Prg->GlobalData[0x28],int64to32(WrittenFileSize>>32));
     VM.Execute(Prg);
   }
 }
